/*
///////////////////////////////////////////////////////////////////////////////////

    Partial STM32 library for demo for the STM32 Blue Pill
    Blog Part 4

    Copyright (C) 2021 - Martin Kaselis (modelrailmusings.weebly.com)

///////////////////////////////////////////////////////////////////////////////////
*/

#include "stm32.h"

#include <stdarg.h>

///////////////////////////////////////////////////////////////////////////////////////////

STM_FLASH *FLASH  = (STM_FLASH*) 0x40022000;
STM_RCC   *RCC    = (STM_RCC*)   0x40021000;
STM_STK   *STK    = (STM_STK*)   0xe000e010;
STM_NVIC  *NVIC   = (STM_NVIC*)  0xe000e100;
STM_GPIO  *GPIOA  = (STM_GPIO*)  0x40010800;
STM_GPIO  *GPIOB  = (STM_GPIO*)  0x40010c00;
STM_GPIO  *GPIOC  = (STM_GPIO*)  0x40011000;
STM_ADC   *ADC1   = (STM_ADC*)   0x40012400;
STM_USART *USART1 = (STM_USART*) 0x40013800;

///////////////////////////////////////////////////////////////////////////////////////////

// Temporary crash routine, will tie to serial output and assert() later on.
void crash( void )
{
    gpioInit();

    gpioPinInit( GPIO_PC13, GPIO_MODE_OUTPUT );

    for(;;)
    {
        gpioPinWrite( GPIO_PC13, 1 );
        delayUsec(50*1000);
        gpioPinWrite( GPIO_PC13, 0 );
        delayUsec(50*1000);
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Initialise the system clock to the HSE at 72MHz, leave HSI running since some peripherals still need it.
void clockInit( void )
{
    // Conf clock : 72MHz using HSE 8MHz crystal w/ PLL X 9 (8MHz x 9 = 72MHz)
    FLASH->ACR = ( FLASH->ACR & ~FLASH_ACR_LATENCY_MASK ) | FLASH_ACR_LATENCY_2;      // Two wait states, per datasheet

    RCC->CFGR       |= RCC_CFGR_PPRE1_2;    // prescale AHB1 = HCLK/2
    RCC->CR         |= RCC_CR_HSEON;        // enable HSE clock
    while( !(RCC->CR & RCC_CR_HSERDY) )     // wait for the HSEREADY flag
        ;

    RCC->CFGR       |= RCC_CFGR_PLLSRC | RCC_CFGR_PLLMULL9; // set PLL source to HSE and multiply by 9
    RCC->CR         |= RCC_CR_PLLON;        // enable the PLL
    while( !(RCC->CR & RCC_CR_PLLRDY) )     // wait for the PLLRDY flag
        ;

    RCC->CFGR       |= RCC_CFGR_SW_PLL;     // set clock source to pll
    while( !(RCC->CFGR & RCC_CFGR_SWS_PLL) )   // wait for PLL to be CLK
        ;
}

// Delay N microseconds
void delayUsec( uint32_t usec )
{
    uint32_t i;

    // approximate calibration for 72MHz clock
    for( i = 0; i < 4 * usec; i++ )
        ;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Init systick subsystem
void systickInit( void )
{
    STK->RVR = 72000-1; // 1ms at 72MHz
    STK->CVR = 0;
    STK->CSR = STK_CSR_PROCESSOR_CLK | STK_CSR_INTERRUPT | STK_CSR_ENABLE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

void nvicInit( void )
{
    // Nothing to do, but maintain a consistent interface
}

// Enable a peripheral interrupt
void nvicEnable( int n )
{
    uint32_t *pReg = (uint32_t*) &NVIC->ISER[ n / 32 ];

    *pReg |= 1 << ( n & 31 );
}

// Disable a peripheral interrupt
void nvicDisable( int n )
{
    uint32_t *pReg = (uint32_t*) &NVIC->ICER[ n / 32 ];

    *pReg |= 1 << ( n & 31 );
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Init the GPIO subsystem
void gpioInit(void)
{
    // RCC_APB2ENR: Enable I/O port C, B and A clocks, also AFIO clock for PB3/PB4 remapping (not used here)
    RCC->APB2ENR |= IOPC_EN | IOPB_EN | IOPA_EN | AFIOEN;
}

// Init one GPIO pin
void gpioPinInit( GPIO_BIT bit, GPIO_MODE mode )
{
    uint32_t *pReg;
    uint32_t shift = ( bit & 7 ) * 4;

    if(      bit <= GPIO_PA7  ) pReg = (uint32_t *) &GPIOA->CRL;
    else if( bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->CRH;
    else if( bit <= GPIO_PB7  ) pReg = (uint32_t *) &GPIOB->CRL;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->CRH;
    else if( bit <= GPIO_PC7  ) pReg = (uint32_t *) &GPIOC->CRL;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->CRH;
    else
        crash();

    switch( mode )
    {
    case GPIO_MODE_INPUT:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x04 << shift ); // input, floating (no pullup)
        break;
    case GPIO_MODE_OUTPUT:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x01 << shift ); // output push-pull, max speed 10MHz
        break;
    case GPIO_MODE_ANALOG:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x00 << shift ); // analog input
        break;
    case GPIO_MODE_AF_PUSHPULL:
        *pReg = ( *pReg & ~( 0x0f << shift ) ) | ( 0x09 << shift ); // alternate function push-pull output, 10 Mhz
        break;
    default:
        crash();
        break;
    }
}

// Write multiple bits under a mask.  Must all be in the same bank (A, B or C).  0 bits are cleared first, then 1 bits set.
void gpioPinWriteMultiple( GPIO_BIT bit, uint32_t mask, uint32_t value )
{
    uint32_t *pReg;
    uint32_t data;

    if(      bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->BSRR;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->BSRR;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->BSRR;
    else
        crash();

    // turn 0 bits off, pause, then 1 bits on (protect FETs from H & L being on together)
    data = ( ( ~value & mask ) << 16 );
    *pReg = data;

    delayUsec( 1 );

    data = ( value & mask );
    *pReg = data;
}

// Write one GPIO pin
void gpioPinWrite( GPIO_BIT bit, uint32_t on )    // Takes approx 1usec per call
{
    uint32_t *pReg;
    uint32_t shift = ( bit & 15 );

    if(      bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->BSRR;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->BSRR;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->BSRR;
    else
        crash();

    if( on )
        *pReg = 1 << shift;
    else
        *pReg = 1 << ( shift + 16 );
}

// Read one GPIO pin
uint32_t gpioPinRead( GPIO_BIT bit )
{
    uint32_t *pReg;
    uint32_t shift = ( bit & 15 );

    if(      bit <= GPIO_PA15 ) pReg = (uint32_t *) &GPIOA->IDR;
    else if( bit <= GPIO_PB15 ) pReg = (uint32_t *) &GPIOB->IDR;
    else if( bit <= GPIO_PC15 ) pReg = (uint32_t *) &GPIOC->IDR;
    else
        crash();

    return ( *pReg >> shift ) & 1;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Initialise the ADC1 peripheral
void adcInit( void )
{
    RCC->CFGR       |= RCC_CFGR_ADCPRE_DIV6;   // Must be <= 14MHz, so 72MHz clock / 6 => 12 MHz
    RCC->APB2ENR    |= ADC1EN;  // enable ADC clock

    ADC1->CR2 |= ADC_CR2_ADON; // turn on ADC

    delayUsec( 10 );    // 10us delay between ADC ON and start of calibration

    ADC1->CR2 |= ADC_CR2_RSTCAL; // reset calibration registers
    while( ( ADC1->CR2 & ADC_CR2_RSTCAL ) != 0 ) // wait for calibration registers to be initalized
        ;
    ADC1->CR2 |= ADC_CR2_CAL; // start calibration
    while( ( ADC1->CR2 & ADC_CR2_CAL ) != 0 ) // wait for calibration to complete
        ;
}

// Init an I/O pin as an ADC input, and tie it to a specific ADC channel.
void adcPinInit( GPIO_BIT bit, uint32_t channel )
{
    gpioPinInit( bit, GPIO_MODE_ANALOG );

    ADC1->SMPR2 |= ( ADC_SMPR_28 << 3 * channel ); // set sampling time to 28.5 cycles, or approx 2 us
}

// Read an ADC input, returning the result as 0..4095.
uint32_t adcReadChannel( uint32_t channel )
{
    int i;

    ADC1->SQR3 = ( channel & ADC_SQR_MASK );
    ADC1->SR &= ~ADC_SR_EOC;
    ADC1->CR2 |= ADC_CR2_ADON; // start conversion

    for( i = 0; i < 300; i++ )  // timeout just in case: I'd rather have a bad reading than a lockup.
        if( ( ADC1->SR & ADC_SR_EOC ) != 0 )
            break;

    return ADC1->DR & ADC_DR_MASK;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// USART configured as a standard 38400 interface, TX side only for now, using interrupts and with a printf() function.

#define COMMS_SIZE       255    // must be 2^N-1

static char buffer[ COMMS_SIZE + 1 ];
static int readIndex = 0;               // only updated by the ISR
static int writeIndex = 0;              // only updated by the put function.  Data availabel if write != read; max capacity is COMMS_SIZE-1

// Initialise USART1 as a 38400 baud serial terminal, TX only, on PA9 and PA10.
void usartInit( void )
{
    RCC->APB2ENR |= USART1EN | AFIOEN;  // Need the AFIO, even if we are using the default pins.

    gpioPinInit( GPIO_PA9, GPIO_MODE_AF_PUSHPULL );     // Required pin definitions
    gpioPinInit( GPIO_PA10, GPIO_MODE_INPUT );

    USART1->CR3 = 0; // No features used
    USART1->CR2 = 0; // 1 stop bit
    USART1->BRR = 72000000/38400;       // system clock / baud rate.
    USART1->CR1 = ( USART1->CR1 & ~USART_CR1_CLEAR_MASK ) | USART_CR1_TE | USART_CR1_RE | USART_CR1_UE; // Enable TX, RX and USART

    nvicEnable( NVIC_INT_USART1 );
}

// USART1 interrupt handler; function name already defined in stm32_startup.c
void USART1_IRQHandler( void )
{
    // TXE set whenever teh TX buffer is ready to accept a character
    if( USART1->SR & USART_SR_TXE )
    {
        USART1->DR = buffer[ readIndex ];

        // read should always be NE write here, but allow for hiccups.
        if( readIndex != writeIndex )
            readIndex = ( readIndex + 1 ) & COMMS_SIZE;
    }

    // If that was the last character, mask the interrupt.  TXE will still be set, and the put function will restart things by unmasking again.
    if( readIndex == writeIndex )
        USART1->CR1 &= ~USART_CR1_TXEIE;
}

// Put one character into the TX buffer, and restart interrupts if necessary.
void usartPutChar( char c )
{
    // Discard characters if the buffer overlflows: this port is intended as a debug feature.
    if( ( ( writeIndex - readIndex ) & COMMS_SIZE ) < COMMS_SIZE )
    {
        buffer[ writeIndex ] = c;
        writeIndex = ( writeIndex + 1 ) & COMMS_SIZE;

        // If the ISR is idle, give it a kick...
        if( !( USART1->CR1 & USART_CR1_TXEIE ) )
            USART1->CR1 |= USART_CR1_TXEIE;
    }
}

// Put a string into the TX buffer
void usartPutStr( char *s )
{
    while( *s )
        usartPutChar( *s++ );
}

// Simplified printf() function, handles most common cases: %04d, %-6s, etc.  %hx forces 16-bit output.
void usartPrintf( char *fmt, ... )
{
    uint32_t i;
    uint32_t n = 0;
    uint32_t value;
    char c;
    const char *str = "";
    char buf[ 32 + 1 ];     // Big enough for a 32-bit binary number.
    va_list ap;

    va_start( ap, fmt );

    while( ( c = *fmt++ ) != '\0' )
    {
        if( c == '\\' )
        {
            c = *fmt++;
            if( c == 'n' )
                c = '\n';
            else if( c == 'r' )
                c = '\r';
            else if( c == 'b' )
                c = '\b';
            else if( c == '0' )
                c = '\0';
            usartPutChar( c );
        }
        else if( c != '%' )
        {
            usartPutChar( c );
        }
        else if( *fmt == '%' )
        {
            fmt++;
            usartPutChar( c );
        }
        else
        {
            bool left = false;
            bool zero = false;
            bool negative = false;
            bool upper = false;
            uint32_t width = 0;
            uint32_t radix = 0;
            buf[0] = '\0';

            value = va_arg( ap, uint32_t );

            // Collect width and fill information
            if( *fmt == '-' )
            {
                left = true;
                fmt++;
            }
            if( *fmt == '0' )
            {
                zero = true;
                fmt++;
            }
            while( *fmt >= '0' && *fmt <= '9' )
            {
                width = width * 10 + (uint8_t) *fmt - '0';
                fmt++;
            }

            // Determine the radix and make everything unsigned
            c = *fmt++;
            if( c == 'l' )
                c = *fmt++;
            if( c == 'h' )
            {
                value &= 0xffff;
                c = *fmt++;
            }
            switch( c )
            {
            case 'd':
                radix = 10;
                if( (int32_t) value < 0 )
                {
                    negative = true;
                    value = (uint32_t) ( - (int32_t) value );
                }
                break;
            case 'u': radix = 10; break;
            case 'b': radix = 2;  break;
            case 'x': radix = 16; break;
            case 'X': radix = 16; upper = true; break;
            case 'p': radix = 16; width = 8; zero = true; break;
            case 'c': break;    // handled below
            case 's': break;    // handled below
            default:
                radix = 0;
                c = '0';
                break;
            }

            // Build up the numeric string backwards into buf[]
            i = 0;
            if( c == 'c' )
            {
                buf[ i++ ] = (uint8_t) value;
            }
            else if( c == 's' )
            {
                str = ( value != 0 )? (char*) value : "<null>";
                for( i = 0; i < 100 && str[ i ]; i++ )
                {
                }
            }
            else if( radix == 0 || value == 0 )
            {
                buf[ i++ ] = '0';
            }
            else
            {
                while( value > 0 )
                {
                    uint32_t digit = value % radix;
                    if( digit <= 9 )
                        digit += '0';
                    else if( upper )
                        digit += 'A' - 10;
                    else
                        digit += 'a' - 10;

                    buf[ i++ ] = digit;
                    value /= radix;
                }
            }
            if( negative )
                buf[ i++ ] = '-';

            // Finally, copy the number in correct order, padding before or after as required.
            width = ( width > i )? width - i : 0;
            for( ; !left && width > 0; width-- )
            {
                usartPutChar( zero? '0' : ' ' );
            }
            if( c == 's' )
            {
                for( uint32_t j = 0; j < i; j++ )
                    usartPutChar( str[ j ] );
            }
            else
            {
                for( ; i > 0; i-- )
                    usartPutChar( buf[ i-1 ] );
            }
            for( ; left && width > 0; width-- )
            {
                usartPutChar( ( zero? '0' : ' ' ) );
            }
        }
    }

    va_end( ap );
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// These functions treat the top 1K of the 64K FLASH as a simulated EEPROM.
// The data is accessed as 32-bit words but with byte addresses (0,4,8,...).  This makes sense given how I am planning to use it.

// Initialise the FLASH (simulated EEPROM) module.
void flashInit( void )
{
    // nothing to do, but keep the init function for consistency.
}

// Read 4 bytes from the flash.  Could be done as a direct memory read, but having a matched pair of read/write functions is usually a good idea.
uint32_t flashRead( uint32_t a )
{
    return *(uint32_t*) ( EEPROM_ADDR + a );
}

// Helper function to wait for a status bit, since this is needed a lot in the FLASH code.
static void flashWait( uint32_t flag )
{
    int i;

    if( flag == FLASH_SR_BSY )
    {
        // This is a negative test: wait a short while until NOT BUSY.  Include the timeout so the system doesn't hang if something goes wrong.
        for( i = 0; i < 10000; i++ )
            if( !( FLASH->SR & FLASH_SR_BSY ) )
                break;
    }
    else if( flag == FLASH_SR_EOP )
    {
        // Wait a long while until END OF OPERATION
        for( i = 0; i < 1000000; i++ )
            if( FLASH->SR & FLASH_SR_EOP )
                break;
    }
}

// Erase the whole page of simulated EEPROM.
void flashErase( void )
{
    // The sequence from the manual (PM0075)...
    flashWait( FLASH_SR_BSY );

    if( FLASH->CR & FLASH_CR_LOCK )
    {
        FLASH->KEYR = FLASH_KEYR_KEY1;
        FLASH->KEYR = FLASH_KEYR_KEY2;
    }

    FLASH->CR |= FLASH_CR_PER;  // Page Erase
    FLASH->AR = EEPROM_ADDR;
    FLASH->CR |= FLASH_CR_STRT; // Start the erasure

    flashWait( FLASH_SR_EOP );

    // The part that isn't in the documentation - we have to turn those control bits OFF again!
    FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~( FLASH_CR_PER | FLASH_CR_STRT );

    FLASH->CR |= FLASH_CR_LOCK;
}

// Write one 32-bit word into the simulated EEPROM area.  Addressing is via a byte offset (0,4,8,...).
bool flashWrite( uint32_t a, uint32_t d )
{
    // The sequence from the manual (PM0075)...
    flashWait( FLASH_SR_BSY );

    if( FLASH->CR & FLASH_CR_LOCK )
    {
        FLASH->KEYR = FLASH_KEYR_KEY1;
        FLASH->KEYR = FLASH_KEYR_KEY2;
    }

    // The writes have to be as 16-bit halfwords, so done in two parts
    FLASH->CR |= FLASH_CR_PG;
    *(uint16_t*) ( EEPROM_ADDR + a ) = (uint16_t) d;

    flashWait( FLASH_SR_BSY );

    FLASH->CR |= FLASH_CR_PG;
    *(uint16_t*) ( EEPROM_ADDR + a + 2 ) = (uint16_t) ( d >> 16 );

    flashWait( FLASH_SR_EOP );

    // The part that isn't in the documentation - we have to turn those control bits OFF again!
    FLASH->SR |= FLASH_SR_EOP;
    FLASH->CR &= ~( FLASH_CR_PG );

    return ( *(uint32_t*) ( EEPROM_ADDR + a ) == d ); // TRUE if the write worked
}

// Lock the FLASH again when all accesses have finished.
void flashLock( void )
{
    FLASH->CR |= FLASH_CR_LOCK;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Bit-bashed SPI implementation, running at approx 300kHz.
// Support for two MCP23S17 16-bit I/O expanders, used to drive 32 pairs of DPDT relays.

#define EXPANDER_1      GPIO_PA11       // Chip select lines, arbitrary choices from the few pins I wasn't using for other jobs.
#define EXPANDER_2      GPIO_PC14

static uint32_t lastXpndData;

// Init the SPI module
void spiInit( void )
{
    // Define the pins - the default SPI1 collection except for NSS
    gpioPinInit( GPIO_PA5, GPIO_MODE_OUTPUT );
    gpioPinInit( GPIO_PA6, GPIO_MODE_INPUT );
    gpioPinInit( GPIO_PA7, GPIO_MODE_OUTPUT );

    // Define the two CS pins for the I/O expanders
    gpioPinInit( EXPANDER_1, GPIO_MODE_OUTPUT );
    gpioPinInit( EXPANDER_2, GPIO_MODE_OUTPUT );
    gpioPinWrite( EXPANDER_1, 1 );
    gpioPinWrite( EXPANDER_2, 1 );
}

// Send a string of bytes to one device.  Using an array would have been simpler, and necessary if handling data buffers,
// but this approach keeps the application code simpler for this type of device.  In practice, all transfers are 3 or 4 bytes long.
// Usage: spiSend( device, N, byte1, byte2, ..., byteN );
void spiSend( uint32_t cs, int n, ... )
{
    va_list ap;
    int i, j;
    uint32_t value;

    va_start( ap, n );
    gpioPinWrite( cs, 0 );

    for( i = 0; i < n; i++ )
    {
        value = va_arg( ap, uint32_t );

        // Each call to pinWrite takes about 1us, so each byte takes about 25-30us (300kHz or thereabouts).
        for( j = 0x80; j > 0; j >>= 1 )
        {
            gpioPinWrite( GPIO_PA7, ( value & j ) );
            gpioPinWrite( GPIO_PA5, 1 );
            gpioPinWrite( GPIO_PA5, 0 );
        }
    }

    gpioPinWrite( cs, 1 );
}

// Init the I/O expanders, setting all 16 pins as outputs, and preparing to do all data writes with the full 16 bits.
void xpndInit( void )
{
    uint32_t x;

    for( x = EXPANDER_1; x <= EXPANDER_2; x += EXPANDER_2 - EXPANDER_1 )
    {
        spiSend( x, 3, XPND_COMMAND, XPND_IOCON2, XPND_IOCON_SEQOP );  // just in case the wrong bank mode somehow got set.  See the MCP23S17 datasheet for details.
        spiSend( x, 3, XPND_COMMAND, XPND_IOCON1, XPND_IOCON_SEQOP );
        spiSend( x, 4, XPND_COMMAND, XPND_IODIR,  0x00, 0x00 );        // all pins are outputs
        spiSend( x, 4, XPND_COMMAND, XPND_GPIO,   0x00, 0x00 );        // and all set to 0
    }

    lastXpndData = 0x00000000; // Used to optimise (skip) unneeded writes.
}

// Write a new set of 32 output values to the expanders.  This function can take a maximum of 250us or so to run.
void xpndWriteData( uint32_t value )
{
    // Only write to each device if at least one of its bits has changed.
    if( ( ( value ^ lastXpndData ) & 0xffff ) != 0 )
        spiSend( EXPANDER_1, 4, XPND_COMMAND, XPND_GPIO, (value >>  0) & 0xff, (value >>  8) & 0xff );
    if( ( ( value ^ lastXpndData ) & 0xffff0000 ) != 0 )
        spiSend( EXPANDER_2, 4, XPND_COMMAND, XPND_GPIO, (value >> 16) & 0xff, (value >> 24) & 0xff );

    lastXpndData = value;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
